/*
 *
 *
 *   sgclient.cc
 *
 *   Command-line client to interface with SGD serial gateway daemon.
 *   
 *   Copyright (C) 2003, by the Grand River Dam Authority,
 *   All rights reserved.
 *
 *   Author: Matt Roberts
 *
 *
 */


// Standard C++ includes
#include <iostream>
#include <string>

// my++ includes
#include <my++/unixsocket.h>
#include <my++/select.h>

// Standard C includes
#include <cstring>
#include <cstdio>

// Unix C includes
#include <unistd.h>
#include <errno.h>

// local includes
#include "util.h"
#include "iobuf.h"
#include "signals.h"

// size of buffer to use for read/write calls
#define CBUF_SIZE (4096)

// prototypes
static int ioloop (my::unixsocket &sock, bool readonly = false);


// global data
static unsigned short verbose = 0;


/*
 *
 *   usage(...)
 *
 */

static void usage (std::ostream &out = std::cerr) {
	out << std::endl;
	out << "Usage: sgclient [-rv] <socket_path> [<command> [<arg1> <arg2> ... <argN> ]]"
		<< std::endl << std::endl;
	out << "Where:" << std::endl;
	out << "\t-r   enable readonly mode (stdin active, but ignored)" << std::endl;
	out << "\t-v   increase verbosity (default = 0)" << std::endl;
	out << std::endl;
	exit(111);
}


/*
 *
 *
 *   main (...)
 *
 *   Main function of program.
 *
 *
 */

int main (int argc, char **argv) {
	// readonly flag
	bool readonly = false;
	std::string remote;
	std::string command;
	char * const *args = 0;

	// setup signals
	setup_signals();

	// make sure we have enough arguments
	if (argc < 2)
		usage();
	
	// parse arguments
	for (args = argv + 1; (*args != NULL) && ((*args)[0] == '-'); ++args) {
		for (size_t i = 1; i < strlen(*args); ++i) {
			// DEBUG:
			if (verbose >= 2)
				std::cerr << "Parsing option: " << (*args)[i] << std::endl;

			// select option
			switch ((*args)[i]) {
				case 'v':
					++verbose;
					break;
				case 'r':
					readonly = true;
					break;
				default:
					usage();
			}
		}
	}

	// if there is no path remaining, bail
	if (*args == NULL)
		usage();

	// extract the socket name
	remote = getsockpath(*args++);

	// extract the command, everything else is arguments
	if (*args != NULL) {
		command = *args++;

		if (verbose >= 1)
			std::cerr << "Command is \'" << command << "\'" << std::endl;
	}

	/*
	if (strcmp(argv[1], "-r") == 0) {
		if (argc < 3)
			usage();
		remote = getsockpath(argv[2]);
		if (argc >= 4) {
			command = argv[3];
			args = argv + 3;
		}
		readonly = true;
	} else {
		remote = getsockpath(argv[1]);
		if (argc >= 3) {
			command = argv[2];
			args = argv + 2;
		}
	}
	*/

	// socket object
	my::unixsocket sock;

	try {
		// connect to remote
		sock.connect(remote);
	} catch (const std::exception &e) {
		std::cerr << "Socket error: could not connect to \"" << remote << "\"" << std::endl;
		return 1;
	}

	// if there is no command, then just do an I/O loop
	if (command.empty())
		return ioloop(sock, readonly);

	// setup file descriptors for new process
	::close(0);
	dup2(sock.getfd(), 0);
	if ( ! readonly) {
		::close(1);
		dup2(sock.getfd(), 1);
	}

	// run the command
	execvp(command.c_str(), args);

	// if that failed, complain
	std::cerr << "Could not exec() command \"" << command << "\": " << strerror(errno) << std::endl;

	// should never reach here
	return 111;
}


/*
 *
 *   ioloop(...)
 *
 */

static int ioloop (my::unixsocket &sock, bool readonly) {
	my::fdset rs, ws;

	iobuffer tx_sock_buf;
	iobuffer rx_sock_buf;

	char in_buffer [CBUF_SIZE];
	ssize_t rtn = 0;

	size_t limit;

	for (;;) {
		rs.zero();
		ws.zero();

		if (!rx_sock_buf.full())
			rs.set(sock.getfd());
		if (!tx_sock_buf.full())
			rs.set(STDIN_FILENO);
		if (!tx_sock_buf.empty())
			ws.set(sock.getfd());
		if (!rx_sock_buf.empty())
			ws.set(STDOUT_FILENO);

		// wait for something to happen
		rtn = my::select(&rs, &ws, NULL, NULL);

		// deal with select errors
		if (rtn < 0) {
			if (errno == EINTR) {
				// if any signals occurred, bail
				if (signal_count(SIGHUP))
					return 0;
				if (signal_count(SIGTERM))
					return 0;
				if (signal_count(SIGINT))
					return 0;
				continue;
			} else {
				std::cerr << "Error occurred during select(): " << strerror(errno) << std::endl;
				return true;
			}
		}

		if (rtn == 0)
			continue;

		if (rs.isset(sock.getfd())) {
			try {

				//
				//  READ from SOCKET
				//

				rtn = sock.get(in_buffer, sizeof(in_buffer));

				// if EOF reached, bail
				if (rtn == 0) {
					if (verbose)
						std::cerr << "Socket closed by sgd server" << std::endl;
					return true;
				}

				// put the data in the buffer
				rx_sock_buf.push(in_buffer, rtn);

			} catch (const my::device::exception &e) {
				if (e.errnum() != EINTR) {
					std::cerr << "Error reading from socket: " << e.what() << std::endl;
					return true;
				}
			} catch (const my::unixsocket::exception &e) {
				if (e.errnum() != EINTR) {
					std::cerr << "Error reading from socket: " << e.what() << std::endl;
					return true;
				}
			}
		}

		if (rs.isset(STDIN_FILENO)) {

			//
			//  READ from STDIN
			//

			rtn = read(STDIN_FILENO, in_buffer, sizeof(in_buffer));
			//rtn = read(STDIN_FILENO, tx_sock_buf);

			if (rtn < 0) {
				if (errno != EINTR) {
					std::cerr << "Error while reading from stdin: " << strerror(errno) << std::endl;
					return false;
				}
			} else if (rtn == 0) {// if EOF reached, bail
				if (verbose)
					std::cerr << "Client connection closed" << std::endl;
				return false;
			} else {
				tx_sock_buf.insert(tx_sock_buf.end(), in_buffer, in_buffer + rtn);
			}
		}


		if (ws.isset(sock.getfd())) {
			if (tx_sock_buf.size() == 0) {
				std::cerr << "INTERNAL: Received transmit event on socket, but no data in queue" << std::endl;
			} else {
				try {

					//
					//  write to SOCKET
					//

					if (readonly) {
						// if we are readonly, just pretend to write data :)
						rtn = tx_sock_buf.size();
					} else {
						// else, write the data to the socket
						rtn = sock.put(tx_sock_buf);
					}

					if (rtn > 0)
						tx_sock_buf.erase(tx_sock_buf.begin(), tx_sock_buf.begin() + rtn);

				} catch (const my::device::exception &e) {
					if (e.errnum() != EINTR) {
						std::cerr << "Error writing to socket: " << e.what() << std::endl;
						return true;
					}
				} catch (const my::unixsocket::exception &e) {
					if (e.errnum() != EINTR) {
						std::cerr << "Error writing to socket: " << e.what() << std::endl;
						return true;
					}
				}
			}
		}

		if (ws.isset(STDOUT_FILENO)) {
			if (rx_sock_buf.size() == 0) {
				std::cerr << "INTERNAL: Received transmit event from client, but no data in queue" << std::endl;
			} else {

				//
				//  WRITE to CLIENT
				//

				limit = rx_sock_buf.size();
				if (limit > sizeof(in_buffer))
					limit = sizeof(in_buffer);

				std::copy(rx_sock_buf.begin(), rx_sock_buf.begin() + limit, in_buffer);

				rtn = write(STDOUT_FILENO, in_buffer, rtn);

				if (rtn < 0) {
					if (errno != EINTR) {
						std::cerr << "Error while writing to stdout: " << std::endl;
						return false;
					}
				} else if (rtn == 0) {
					/* noop */
				} else {
					rx_sock_buf.erase(rx_sock_buf.begin(), rx_sock_buf.begin() + rtn);
				}
			}
		}
	}

	return 0;
}

// EOF
