/*
 *
 *
 *   sgconfig.cc
 *
 *   Copyright (C) 2003, by the Grand River Dam Authority,
 *   All rights reserved.
 *
 *   Author: Matt Roberts
 *
 *
 */


// local headers
#include "sgdconfig.h"
#include "util.h"

// my++ headers
#include <my++/cmdline.h>
#include <my++/stype.h>

// C++ headers
#include <iostream>
#include <fstream>
#include <iomanip>

// C headers
#include <stdlib.h>


static const std::string DEFAULT_CONF_FILE  ("/etc/sgd.conf");
static const std::string DEFAULT_PID_FILE   ("/var/run/sgd.pid");
static const std::string DEFAULT_SOCK_PREFIX("/var/run/sgd");

using namespace my;
using namespace std;


/*
 *
 *   usage()
 *
 */

static void usage (std::ostream &out = std::cerr) {
	out << "Usage: sgd <options>" << endl << endl;
	out << "where valid options may include:" << endl;
	out << "\t-c <file>   Specify configuration file (default: " << DEFAULT_CONF_FILE << ")" << endl;
	out << "\t-f          Run in foreground, and log to stderr." << endl;
	out << "\t-s          Disable syslog support." << endl;
	out << endl;
}


/*
 *
 *   islocalpath(...)
 *

static bool islocalpath (const std::string &s) {
	if (s[0] == '/')
		return false;
	if (s.substr(0, 2) == "./")
		return false;
	if (s.substr(0, 3) == "../")
		return false;

	return true;
}
 */


/*
 *
 *   server_config::read_args()
 *
 */

void server_config::read_args (int argc, char **argv) {
	command_line::config cfg;

	cfg.add("-c", command_line::arg_required);
	cfg.add("-f", command_line::arg_none);
	cfg.add("-s", command_line::arg_none);

	command_line::parser args(cfg);

	try {
		args.parse(argc, argv);
	} catch (const std::exception &e) {
		cerr << "Error: " << e.what() << endl << endl;
		usage();
		exit(0);
	}

	if (args.is_set("-c")) {
		o_config = args.get("-c").value();
	}

	if (args.is_set("-f")) {
		o_stderr = true;
		o_daemon = false;
	}

	if (args.is_set("-s")) {
		o_syslog = false;
	}
}

void server_config::read_config (void) {
	string line, keyword, value;
	port_config pc;
	bool port_declared = false;
	size_t line_num = 0;

	ifstream in (o_config.c_str());
	if (!in)
		throw config_error(0, "Could not open config file: " + o_config);

	while (in) {
		if (!getline(in, line))    // get one line
			break;
		++line_num;                // increment line number
		line = rtrim(ltrim(line)); // trim the line
		if (! line.size())         // skip the line if empty
			continue;
		if (line[0] == '#')        // skip if a comment
			continue;

		// read the keyword and value from the line
		keyword = toLower(get_keyword(line));
		value   = get_value (line);

		if ((keyword == "port") || (keyword == "device")) {
			port_declared = true;
			if (!value.size())
				throw config_error(line_num, "Must specify TTY device name");
			if (pc.s_device.size()) {
				if (!pc.s_address.size())
					throw config_error(line_num, "Must specify socket pathname");
				add(pc);
			}
			pc.clear();
			pc.s_device = value;
		} else if (keyword == "owner") {
			if (!port_declared)
				throw config_error(line_num, "Port parameter specified outside port declaration");
			if (!value.size())
				throw config_error(line_num, "Must specify owner name");
			pc.s_owner = value;
		} else if (keyword == "group") {
			if (!port_declared)
				throw config_error(line_num, "Port parameter specified outside port declaration");
			if (!value.size())
				throw config_error(line_num, "Must specify group name");
			pc.s_group = value;
		} else if (keyword == "socket") {
			if (!port_declared)
				throw config_error(line_num, "Port parameter specified outside port declaration");
			if (!value.size())
				throw config_error(line_num, "Must specify socket pathname");
			pc.s_address = getsockpath(value);
		} else if (keyword == "speed") {
			if (!port_declared)
				throw config_error(line_num, "Port parameter specified outside port declaration");
			if (!value.size())
				throw config_error(line_num, "Must specify device speed (bps)");
			pc.s_ios.setSpeed(atoi(value.c_str()));
		} else if (keyword == "parity") {
			if (!port_declared)
				throw config_error(line_num, "Port parameter specified outside port declaration");
			if (!value.size())
				throw config_error(line_num, "Must specify parity (N|E|O)");
			pc.s_ios.setParity(value[0]);
		} else if (keyword == "mode") {
			if (!port_declared)
				throw config_error(line_num, "Port parameter specified outside port declaration");
			if (!value.size())
				throw config_error(line_num, "Must specify valid octal mode");
			pc.s_mode = strtol(value.c_str(), NULL, 8);
		} else if ((keyword == "word-length") || (keyword == "data-bits")) {
			if (!port_declared)
				throw config_error(line_num, "Port parameter specified outside port declaration");
			if (!value.size())
				throw config_error(line_num, "Must specify word length (7|8)");
			pc.s_ios.setCharSize(atoi(value.c_str()));
		} else if (keyword == "stop-bits") {
			if (!port_declared)
				throw config_error(line_num, "Port parameter specified outside port declaration");
			if (!value.size())
				throw config_error(line_num, "Must specify stop bits (1|2)");
			pc.s_ios.setStopBits(atoi(value.c_str()));
		} else if (keyword == "hard-flow") {
			if (!port_declared)
				throw config_error(line_num, "Port parameter specified outside port declaration");
			if (!value.size())
				throw config_error(line_num, "Boolean parameter requires argument");
			switch (yes_or_no(value)) {
				case 0: pc.s_ios.setHWflow(false); break;
				case 1: pc.s_ios.setHWflow(true); break;
				case -1: throw config_error(line_num, "Boolean parameter requires yes/no/true/false argument"); break;
				default: break;
			}
		} else if (keyword == "soft-flow") {
			if (!port_declared)
				throw config_error(line_num, "Port parameter specified outside port declaration");
			if (!value.size())
				throw config_error(line_num, "Boolean parameter requires argument");
			switch (yes_or_no(value)) {
				case 0: pc.s_ios.setSWflow(false); break;
				case 1: pc.s_ios.setSWflow(true); break;
				case -1: throw config_error(line_num, "Boolean parameter requires yes/no/true/false argument"); break;
				default: break;
			}
		} else if (keyword == "daemon") {
			if (port_declared)
				throw config_error(line_num, "Global parameter specified within port declaration");
			if (!value.size())
				throw config_error(line_num, "Boolean parameter requires argument");
			switch (yes_or_no(value)) {
				case 0: o_daemon = false; break;
				case 1: o_daemon = true; break;
				case -1: throw config_error(line_num, "Boolean parameter requires yes/no/true/false argument"); break;
				default: break;
			}
		} else if (keyword == "stderr") {
			if (port_declared)
				throw config_error(line_num, "Global parameter specified within port declaration");
			if (!value.size())
				throw config_error(line_num, "Boolean parameter requires argument");
			switch (yes_or_no(value)) {
				case 0: o_stderr = false; break;
				case 1: o_stderr = true; break;
				case -1: throw config_error(line_num, "Boolean parameter requires yes/no/true/false argument"); break;
				default: break;
			}
		} else if (keyword == "syslog") {
			if (port_declared)
				throw config_error(line_num, "Global parameter specified within port declaration");
			if (!value.size())
				throw config_error(line_num, "Boolean parameter requires argument");
			switch (yes_or_no(value)) {
				case 0: o_syslog = false; break;
				case 1: o_syslog = true; break;
				case -1: throw config_error(line_num, "Boolean parameter requires yes/no/true/false argument"); break;
				default: break;
			}
		} else if (keyword == "pidfile") {
			if (port_declared)
				throw config_error(line_num, "Global parameter specified within port declaration");
			if (!value.size())
				throw config_error(line_num, "String parameter requires argument");
			o_pidfile = rtrim(ltrim(value));
		} else {
			throw config_error(line_num, "unknown configuration directive");
		}
	}

	if (pc.s_device.size()) {
		if (!pc.s_address.size())
			throw config_error(line_num, "Must specify socket filename");
		add(pc);
	}

	//if (islocalpath(pc.s_device))
	//	pc.s_device = DEFAULT_SOCK_PREFIX + pc.s_device;

	if (! s_ports.size())
		throw config_error(line_num + 1, "No serial ports defined!");

}


/*
 *
 *   server_config::server_config(int, char**)
 *
 */

server_config::server_config(int argc, char **argv)
	: o_config(DEFAULT_CONF_FILE), o_pidfile(DEFAULT_PID_FILE),
	  o_daemon(true), o_stderr(false), o_syslog(true) {

	// read command line arguments
	read_args(argc, argv);

	try {
		// read the configuration file
		read_config();
	} catch (const config_error &e) {
		if (e.line() > 0)
			cerr << o_config << ", line " << e.line() << ": ";
		cerr << e.what() << endl;
		exit(5);
	}

	// reread command line arguments, to allow opts to override config file
	read_args(argc, argv);
}


/*
 *
 *   server_config::server_config(...)
 *
 */

server_config::server_config (const server_config &c)
	: o_config(c.o_config), o_daemon(c.o_daemon),
	  o_stderr(c.o_stderr), o_syslog(c.o_syslog) {
	s_ports.clear();
	s_ports.assign(c.s_ports.begin(), c.s_ports.end());
}


/*
 *
 *   port_config::port_config ()
 *
 */

port_config::port_config (void) {
	clear();
}


/*
 *
 *   port_config::clear()
 *
 */

void port_config::clear (void) {
	s_address.erase();
	s_owner.erase();
	s_group.erase();
	s_mode = 0777;
	s_device.erase();
	max_tx_queue = 0;
	client_limit = 0;
	client_timeout = 0;
	always_read = true;
}


/*
 *
 *   port_config::port_config(...)
 *
 */

port_config::port_config (const port_config &c) {
	s_address      = c.s_address;
	s_owner        = c.s_owner;
	s_group        = c.s_group;
	s_mode         = c.s_mode;
	s_device       = c.s_device;
	s_ios          = c.s_ios;
	max_tx_queue   = c.max_tx_queue;
	client_timeout = c.client_timeout;
	client_limit   = c.client_limit;
	always_read    = c.always_read;
}


/*
 *
 *   port_config::dump(...)
 *
 */

void port_config::dump (ostream &o) {
	o << "Device: " << s_device << endl;
	o << "\tSocket: " << s_address << endl;
	o << "\tOwner:  " << s_owner << endl;
	o << "\tGroup:  " << s_group << endl;
	o << "\tMode:   " << oct << s_mode << dec << endl;
	#if 0
	o << "\tSpeed:  " << s_ios.getSpeed() << endl;
	o << "\tLength: " << s_ios.getCharSize() << endl;
	o << "\tParity: " << (int)s_ios.getParity() << endl;
	o << "\tStop:   " << (int)s_ios.getStopBits() << endl;
	#endif
}
