/*
 *
 *
 *   termapp.cc
 *
 *   Source code for TerminalApplication class common to the different
 *   SG-type terminals.
 *
 *   Copyright (C) 2003, by the Grand River Dam Authority,
 *   All rights reserved.
 *
 *   Author: Matt Roberts
 *
 *
 */


#include <string>
#include <deque>
#include <list>

#include <my++/unixsocket.h>
#include <my++/tty.h>
#include <my++/select.h>

#include <my++/termios.h>
#include <my++/serialios.h>
#include <my++/conios.h>

#include <cstring>
#include <cstdio>

#include <time.h>
#include <signal.h>
#include <curses.h>

#include "signals.h"
#include "rtwin.h"
#include "util.h"
#include "termapp.h"


// let's use std::
using namespace std;

// flags used to open the console tty
const static int default_flags = my::device::not_ctl_tty | my::device::nonblocking;

// keyboard echo OFF
const static bool do_echo = false;

// character to switch to/from read-only mode
const static int ro_key   = KEY_F(1);  // F1 to change mode; 0 = disable

// kill character
const static int kill_key = KEY_F(2);  // F2 to exit; 0 = disable

// debug dump character
const static int debug_key = KEY_F(3); // 0 = disable

// allow SIGINT to kill program
const static bool allow_sigint = false;


/*
 *
 *   signal handler for SIGWINCH signal
 *
 */

static void sig_winch (int signo) {
	/*
	int y, x;
	getmaxyx(stdscr, &y, &x);
	clear();
	resizeterm(y, x);
	*/
}


/*
 *
 *   TerminalApplication::TerminalApplication(...)
 *
 */

TerminalApplication::TerminalApplication(const std::string &tty_dev, WinBase &w_obj)
	: console(tty_dev, default_flags),
	  win_obj(w_obj),
	  termwin(NULL),
	  statwin(NULL),
	  rootscr(NULL),
	  /* ten minutes default R/W timeout */
	  rw_timeout(600) {

	my::termios console_config;
	console_config.makeRaw();
	console.set_options(console_config);

	// setup CURSES
	rootscr = newterm(
			NULL,
			fdopen(console.getfd(), "w"),
			fdopen(console.getfd(), "r")
			); 
	set_term(rootscr);
	cbreak();
	noecho();
	nonl();
	intrflush(stdscr, FALSE);
	keypad(stdscr, TRUE);

	if (has_colors()) {
		// enable color support
		start_color();

		// Simple color assignment
		init_pair(COLOR_BLACK, COLOR_BLACK, COLOR_BLACK);
		init_pair(COLOR_GREEN, COLOR_GREEN, COLOR_BLACK);
		init_pair(COLOR_RED, COLOR_RED, COLOR_BLACK);
		init_pair(COLOR_CYAN, COLOR_CYAN, COLOR_BLACK);
		init_pair(COLOR_WHITE, COLOR_WHITE, COLOR_BLACK);
		init_pair(COLOR_MAGENTA, COLOR_MAGENTA, COLOR_BLACK);
		init_pair(COLOR_BLUE, COLOR_BLUE, COLOR_BLACK);
		init_pair(COLOR_YELLOW, COLOR_YELLOW, COLOR_BLACK);
	}

	// get sizes
	getmaxyx(stdscr, rows, cols);

	// configure individual window sizes
	termwin = subwin(stdscr, rows - 1, 0,        0, 0);
	statwin = subwin(stdscr, 1,        0, rows - 1, 0);

	// set the ris alarm object to own the terminal window
	win_obj = termwin;

	// setup window properties
	idlok(termwin, TRUE);
	idlok(statwin, FALSE);
	scrollok(termwin, TRUE);
	scrollok(statwin, FALSE);

	// use REVERSE video for the status bar
	wattron(statwin, A_REVERSE);

	// clear both windows
	wclear(statwin);
	wclear(termwin);

	// go home and redraw the screen
	setsyx(0, 0);
	refresh();
}


/*
 *
 *  TerminalApplication::~TerminalApplication()
 *
 */

TerminalApplication::~TerminalApplication(void) {
	// clear the screen and cleanup
	wclear(stdscr);
	endwin();

	// close the TTY device
	console.close();

	// delete any window resources used
	if (termwin)
		delwin(termwin);
	if (statwin)
		delwin(statwin);
	if (rootscr)
		delscreen(rootscr);
}


void TerminalApplication::saveCursor(void) {
	int x, y;
	getsyx(y, x);
	c_stack.push(x, y);
}


void TerminalApplication::restoreCursor(void) {
	int x = -1, y = -1;

	c_stack.pop(x, y);

	if ((x >= 0) && (y >= 0))
		setsyx(y, x);
}


void TerminalApplication::status_line (
		const std::string &l_msg,
		const std::string &sock,
		bool read_only) {
	std::string displayed = " " + l_msg;
	std::string path_disp = " | " + path_file(sock) + ((read_only) ? "(ro)" : "(RW)");
	std::string time_disp = " | " + gettime() + " ";
	std::string legend;

	if (read_only)
		legend = " | F1-Enable TX   F2-Exit";
	else
		legend = " | F1-Disable TX  F2-Exit";

	size_t count = cols - (2 + time_disp.size() + path_disp.size() + legend.size());

	// build the status line text
	if (displayed.size() > count)
		displayed = std::string(displayed.begin(), displayed.begin() + count);
	if (displayed.size() < count)
		displayed += std::string(count - displayed.size(), ' ');
	displayed += legend;
	displayed += path_disp;
	displayed += time_disp;

	// save the cursor position and hide it
	saveCursor();
	hideCursor();

	// write the status line
	mvwaddstr(statwin, 0, 0, displayed.c_str());

	// restore the cursor position
	restoreCursor();

	// update both windows
	wrefresh(statwin);
	wrefresh(termwin);

	// show the cursor again
	showCursor();
}


int TerminalApplication::run(const std::string &remote_address) {
	unsigned char buffer [128];
	time_t now = 0, last_update = 0, last_keystroke = 0;
	ssize_t count;
	ssize_t rtn;
	bool read_only = true, stop = false;
	my::fdset rset, wset;
	timeval timeout;
	std::string tx_buffer;

	// setup signal handling
	setup_signals();
	signal(SIGWINCH, sig_winch);

	// start with current timers
	now = last_update = last_keystroke = time(NULL);

	// connection loop
	for (;;) {
		message  = "Connecting to device \'";
		message += remote_address;
		message += "\': ";
		status_line(message, remote_address, read_only);

		rset.set(console.getfd());

		try {
			// connect to socket
			sock.connect(remote_address);
		} catch (const std::exception &e) {
			timeout.tv_sec = 2;
			timeout.tv_usec = 0;

			message  = "Connect failed: ";
			message += e.what();
			message += ".  Trying again in a few moments.";
			status_line(message, remote_address, read_only);

			// watch for I/O
			rtn = select(&rset, &wset, NULL, &timeout);

			if (rtn > 0) {
				if (rset.isset(console.getfd())) {

					//
					//  read data from CONSOLE
					//

					int ch = getch();

					// if EOF
					if (ch == ERR) {
						stop = true;
						return 0;
					}

					// control or function keys
					if (ch > 128) {
						if (kill_key && (ch == kill_key))
							return 0;

						if (debug_key && (ch == debug_key))
							win_obj.test();
					}
				}
			}

			if (allow_sigint && signal_count(SIGINT))
				return 0;
			if (signal_count(SIGTERM))
				return 0;
			continue;
		}

		read_only = true;
		status_line(message, remote_address, read_only);

		rset.set(sock.getfd());

		//
		// I/O loop
		//

		for (stop = false; ! stop ;) {
			timeout.tv_sec = 1;
			timeout.tv_usec = 0;

			// watch for write ok only if data available
			if (tx_buffer.size() > 0) {
				wset.set(sock.getfd());
			} else {
				wset.clear(sock.getfd());
			}

			// watch for I/O
			rtn = select(&rset, &wset, NULL, &timeout);

			// get time
			now = time(NULL);

			// update status line
			if (now > last_update) {
				if (rw_timeout && (! read_only) && ((now - last_keystroke) > rw_timeout)) {
					read_only = ! read_only;
					status_line("Online", remote_address, read_only);
				} else {
					status_line("Online", remote_address, read_only);
					last_update = now;
				}
			}

			if (signal_count(SIGTERM) || signal_count(SIGHUP))
				return 0;

			if (rtn > 0) {
				if (rset.isset(console.getfd())) {

					//
					//  read data from CONSOLE
					//

					int ch = getch();

					// if EOF
					if (ch == ERR) {
						stop = true;
						return 0;
					}

					last_keystroke = time(NULL);

					// control or function keys
					if (ch > 128) {
						if (ro_key && (ch == ro_key)) {
							read_only = ! read_only;
							status_line("Online", remote_address, read_only);
						}

						if (kill_key && (ch == kill_key))
							return 0;

						if (debug_key && (ch == debug_key))
							win_obj.test();
					} else {

						if (! read_only)
							tx_buffer += (char)(ch & 0x00FF);

						if (do_echo) {
							win_obj.put(ch);
							win_obj.refresh();
						}
					}
				}

				if (rset.isset(sock.getfd())) {

					//
					//  read data from SOCKET
					//

					count = sock.get(buffer, sizeof(buffer));

					if (count == 0) {
						stop = true;
						continue;
					}

					if (count > 0) {
						win_obj.write(buffer, count);
						win_obj.refresh();
					}
				}

				if (wset.isset(sock.getfd()) && (tx_buffer.size())) {
					count = sock.put(tx_buffer.c_str(), tx_buffer.size());
					if ((size_t)count == tx_buffer.size())
						tx_buffer.erase();
					else
						tx_buffer.erase(tx_buffer.begin(), tx_buffer.begin() + count);
				}
			}
		}

		// close the socket
		sock.close();

		// complain
		message = "Connection broken, standby...";
		status_line(message, remote_address, read_only);
		sleep(2);
	}

	// should never reach here
	return 0;
}


// EOF
